
from izisat.misc.raster_processing import RasterProcessing        
from izisat.misc.connections import Connections
from izisat.misc.utils import Utils
from izisat.misc.files import Files
from loguru import logger
import geopandas as gpd
from datetime import datetime, timedelta
from shapely.geometry import box


class IZISentinel:
    def __init__(self, output_base_path, username, password):
        """Initializes the Sentinel2_Band_Downloader instance."""
        self.output_base_path = output_base_path
        self.access_token, self.refresh_token, self.dt_access_token = self.connect_to_api(username, password)
        
    def connect_to_api(self, username, password):
        """Connects to the Sentinel API and obtains an access token."""
        connections = Connections()
        access_token, refresh_token, dt_access_token = connections.access_token(username, password)
        return access_token, refresh_token, dt_access_token
                                      
    def construct_query(self, geodataframe, end_date, cloud_cover_percentage, type, platform_name):
        """Constructs a query for retrieving Sentinel-2 products based on specified parameters."""
        utils = Utils()
        start_date = (end_date - timedelta(days=1)).strftime('%Y-%m-%d')
        end_date = end_date.strftime('%Y-%m-%d')
        projected_gdf = gdf.to_crs(geodataframe.estimate_utm_crs('SIRGAS 2000'))
        geodataframe = projected_gdf.buffer(1000)
        geodataframe = geodataframe.to_crs(4674)
        minx, miny, maxx, maxy = geodataframe.total_bounds
        bbox_geom = box(minx, miny, maxx, maxy)
        bbox_gdf = gpd.GeoDataFrame(geometry=[bbox_geom], crs=geodataframe.crs)
        footprint = bbox_gdf.iloc[0].geometry.wkt
        query = utils.construct_query_for_sentinel2_products(footprint, start_date, end_date, cloud_cover_percentage, type, platform_name)
        return query
        
    def products_from_sentinel_2(self, params):
        """Retrieves Sentinel-2 products based on the provided query parameters."""
        connections = Connections()
        products = connections.retrieve_sent_prod_from_query(params)
        return products
    
    def get_products_info(self, products):
        """Retrieves information about Sentinel-2 products."""
        utils = Utils()
        products_info = utils.retrieve_products_info(products)
        return products_info
    
    def output_folder(self, products_info, bands_dict):
        """Creates output folders to save downloaded bands."""
        files = Files()
        directory_paths = files.create_output_folders(self.output_base_path, products_info, bands_dict)
        return directory_paths
    
    def get_bands_links(self, products_info, bands_dict):
        """Retrieves links to bands for Sentinel-2 products."""
        connections = Connections()
        bands_links = connections.retrieve_bands_links(self.access_token, products_info, bands_dict)
        return bands_links
    
    def download_band(self, products_info, bands_link, base_dir, tile):
        """Downloads bands for Sentinel-2 products based on the provided links."""
        connections = Connections()
        connections.download_bands(self.access_token, products_info, bands_link, base_dir, self.dt_access_token, self.refresh_token, tile)

    def download_sentinel2_bands(self, products, bands_dict, tile):
        """Orchestrates the download process for Sentinel-2 bands."""
        utils = Utils()
        if products is None:
            logger.warning("Stopping further execution.")
        else:
            products_info = self.get_products_info(products)
            self.output_folder(products_info, bands_dict)
            links = self.get_bands_links(products_info, bands_dict)
            products=self.download_band(products_info, links, self.output_base_path, tile)

            downloaded_dict = utils.generate_sentinel2_band_paths_nested_by_tile(self.output_base_path, products_info)
            return downloaded_dict
        
    def get_composite_data(self, gdf, downloaded_dict, output_rgb=None, output_ndvi=None, output_evi=None):
        raster_processing = RasterProcessing()
        first_key = next(iter(downloaded_dict))
        first_entry = downloaded_dict[first_key]
        if len(first_entry) == 1:
            band_paths = first_entry[0]
            b02_array, b03_array, b04_array, b08_array, profile = raster_processing.get_cropped_bands_no_merge(gdf, band_paths)
        elif len(first_entry) > 1:
            b02_array, b03_array, b04_array, b08_array, profile = raster_processing.get_cropped_bands_with_merge(gdf, first_entry)

        if output_rgb is not None:
            raster_processing.create_rgb(b02_array, b03_array, b04_array, profile, output_rgb)

        if output_ndvi is not None:
            raster_processing.create_ndvi(b04_array, b08_array, profile, output_ndvi)

        if output_evi is not None:
            raster_processing.create_evi(b04_array, b04_array, b08_array, profile, output_evi)

    def get_satellite_image(self, geodataframe, end_date, cloud_cover_percentage, type='L2A', platform_name='SENTINEL-2', bands_dict={"L2A":{"10m": ["B02", "B03","B04", "B08"]}}, output_rgb=None, output_ndvi=None, output_evi=None):
        query = self.construct_query(geodataframe, end_date, cloud_cover_percentage, type, platform_name)
        products = self.products_from_sentinel_2(query)
        if products == None:
            return
        downloaded_dict = self.download_sentinel2_bands(products, bands_dict, None)
        self.get_composite_data(geodataframe, downloaded_dict, output_rgb, output_ndvi, output_evi)


if __name__ == '__main__':
    DOWNLOAD_FOLDER = '/home/ppz/Documentos/coding/izisat/auxiliary'
    MAX_CLOUD_COVER = 99
    BANDS_DICT = {"L2A":{"10m": ["B02", "B03","B04", "B08"]}}
    PLATFORM_NAME = "SENTINEL-2"
    SATELLITE_TYPE = 'L2A'
    COPERNICUS_USER=''
    COPERNICUS_PASSWD='!'

    zona_unica=False

    ###TESTE ZONA UNICA###
    downloader = IZISentinel(output_base_path=DOWNLOAD_FOLDER, username=COPERNICUS_USER, password=COPERNICUS_PASSWD)
    if zona_unica is True:
        today = datetime(2025, 4, 14, 1, 0, 0)
        gdf = gpd.read_file('/home/ppz/Documentos/coding/izisat/auxiliary/base_ldcelulose/base_cadastral/FULL_20250709_VIEW.shp')
        gdf = gdf.loc[(gdf['TALHAO'] == '0112') & (gdf['ID_PROJETO'] == '5822')]

        output_rgb = '/home/ppz/Documentos/coding/izisat/auxiliary/raster_lake/rgb/teste_rgb.tif'
        output_ndvi = '/home/ppz/Documentos/coding/izisat/auxiliary/raster_lake/ndvi/teste_ndvi.tif'
        downloader.get_satellite_image(
                           geodataframe=gdf,
                           end_date=today,
                           cloud_cover_percentage=MAX_CLOUD_COVER,
                           type=SATELLITE_TYPE,
                           platform_name=PLATFORM_NAME,
                           bands_dict=BANDS_DICT,
                           output_rgb=output_rgb,
                           output_ndvi=output_ndvi,
                           )

    elif zona_unica is not True:
        today = datetime(2025, 4, 9, 1, 0, 0)
        gdf = gpd.read_file('/home/ppz/Documentos/coding/izisat/auxiliary/base_ldcelulose/base_filtrada.geojson')
        gdf = gdf.dissolve(by='PROJETO').reset_index()    
        #gdf = gpd.read_file('/home/ppz/Documentos/coding/izisat/testing.geojson')
        for index, row in gdf.iterrows():
            print(f'linha - {index} de {len(gdf)} -- TALHAO: {row.TALHAO}')
            row_gdf = gdf.iloc[[index]]
            output_rgb = f'/home/ppz/Documentos/coding/izisat/auxiliary/raster_lake/rgb/teste_rgb_{row.TALHAO}.tif'
            output_ndvi = f'/home/ppz/Documentos/coding/izisat/auxiliary/raster_lake/ndvi/teste_ndvi_{row.TALHAO}.tif'
            output_evi = f'/home/ppz/Documentos/coding/izisat/auxiliary/raster_lake/evi/teste_evi_{row.TALHAO}.tif'
            downloader.get_satellite_image(
                            geodataframe=row_gdf,
                            end_date=today,
                            cloud_cover_percentage=MAX_CLOUD_COVER,
                            type=SATELLITE_TYPE,
                            platform_name=PLATFORM_NAME,
                            bands_dict=BANDS_DICT,
                            output_rgb=output_rgb,
                            output_ndvi=output_ndvi,
                            output_evi=output_evi)