import os
import rasterio
import rasterio.mask
from shapely import wkt
from shapely.geometry import mapping
from typing import Union
from pathlib import Path
import geopandas as gpd
import numpy as np
from rasterio.crs import CRS
from rasterio.transform import from_origin
from rasterio.merge import merge
import numpy as np
import rasterio
from rasterio.io import MemoryFile
from rasterio.warp import reproject, Resampling, calculate_default_transform


class RasterProcessing:
    def __init__(self):
        pass

    def crop_band_by_footprint_wkt(self,
                                    jp2_path: str,
                                    footprint_wkt: str,
                                    output_path: str = None
                                ):
        """
        Crops a georeferenced JP2 image using a WKT footprint (MULTIPOLYGON).

        Args:
            jp2_path (str): Path to the .jp2 image (satellite band).
            footprint_wkt (str): WKT string of the footprint polygon (e.g. 'MULTIPOLYGON (...)').
            output_path (str, optional): Output .tif file path or directory. If None, nothing is saved.
            suffix (str): Suffix to append to the output file name if saving to a directory.

        Returns:
            dict: {
                "image": Cropped image as a NumPy array,
                "profile": Raster metadata dictionary,
                "output_name": Full output path if saved, else None
            }
        """
        # Convert WKT string to geometry
        geometry = [mapping(wkt.loads(footprint_wkt))]

        with rasterio.open(jp2_path) as src:
            cropped_image, transform = rasterio.mask.mask(src, geometry, crop=True)
            cropped_image = cropped_image.squeeze()
            profile = src.profile.copy()
            profile.update({
                "height": cropped_image.shape[0],
                "width": cropped_image.shape[1],
                "transform": transform
            })
        if output_path is not None:
            with rasterio.open(output_path, "w", **profile) as dst:
                dst.write(cropped_image)
        return cropped_image, profile

    def ensure_array_in_epsg(self, array: np.ndarray, profile: dict, target_epsg: int):
        import numpy as np
        import rasterio
        from rasterio.io import MemoryFile
        from rasterio.warp import reproject, Resampling, calculate_default_transform
        from rasterio.transform import array_bounds
        """
        Garante que o array raster esteja no EPSG desejado. Se necessário, reprojeta.

        Parâmetros:
        - array: np.ndarray (2D)
        - profile: dict com as chaves: 'transform', 'crs', 'width', 'height'
        - target_epsg: int

        Retorna:
        - dst_array: np.ndarray reprojetado ou original
        - new_profile: dict atualizado
        """
        src_crs = profile.get("crs")
        dst_crs = rasterio.crs.CRS.from_epsg(target_epsg)

        if src_crs == dst_crs:
            new_profile = profile.copy()
            new_profile["crs"] = dst_crs
            return array, new_profile

        # Calcula os bounds a partir do transform
        bounds = array_bounds(profile["height"], profile["width"], profile["transform"])

        transform, width, height = calculate_default_transform(
            src_crs, dst_crs,
            profile["width"], profile["height"],
            *bounds
        )

        dst_array = np.empty((height, width), dtype=array.dtype)

        reproject(
            source=array,
            destination=dst_array,
            src_transform=profile["transform"],
            src_crs=src_crs,
            dst_transform=transform,
            dst_crs=dst_crs,
            resampling=Resampling.nearest
        )

        new_profile = profile.copy()
        new_profile.update({
            "crs": dst_crs,
            "transform": transform,
            "width": width,
            "height": height
        })

        return dst_array, new_profile

    def get_cropped_bands_no_merge(self, gdf:gpd.GeoDataFrame, band_paths):
        utm_zones = {
            19:31979,
            20:31980,
            21:31981,
            22:31982,
            23:31983,
            24:31984,
            25:31985,
        }
        bands_list = ['B02','B03','B04','B08']
        cropped_bands = []

        epsg = utm_zones[int(os.path.basename(band_paths['B02']).split('_')[0][1:3])]
        gdf = gdf.to_crs(epsg)
        footprint_wkt = gdf.iloc[0].geometry.wkt
        
        for band in bands_list:
            path = band_paths.get(band)
            if path:
                cropped, profile = self.crop_band_by_footprint_wkt(path, footprint_wkt)
                cropped_bands.append(cropped)
            
        return (*cropped_bands, profile)

    def get_cropped_bands_with_merge(self, gdf: gpd.GeoDataFrame, band_dicts: list):
        """
        Recorta e mescla múltiplas bandas de entrada organizadas como uma lista de dicionários.
        
        Parâmetros:
        - gdf: GeoDataFrame com a geometria para recorte.
        - band_dicts: lista de dicionários, cada um contendo os caminhos das bandas 'B02', 'B03', 'B04' e 'B08'.
        
        Retorna:
        - Tupla com arrays mesclados para cada banda: B02, B03, B04, B08 e o profile do merge.
        """

        utm_zones = {
            19: 31979,
            20: 31980,
            21: 31981,
            22: 31982,
            23: 31983,
            24: 31984,
            25: 31985,
        }
        bands_list = ['B02', 'B03', 'B04', 'B08']
        cropped_bands_by_type = {band: [] for band in bands_list}
        final_profile = None
        final_crs = gdf.estimate_utm_crs('SIRGAS 2000').to_epsg()

        # Recorta todas as bandas e armazena por tipo
        for band_paths in band_dicts:
            try:
                for band in bands_list:
                    path = band_paths.get(band)
                    if path:
                        epsg = utm_zones[int(os.path.basename(path).split('_')[0][1:3])]
                        gdf = gdf.to_crs(epsg)
                        footprint_wkt = gdf.iloc[0].geometry.wkt
                        cropped, profile = self.crop_band_by_footprint_wkt(path, footprint_wkt)
                        cropped_bands_by_type[band].append((cropped, profile))
            except:
                pass

        # Faz merge por banda
        merged_bands = []
        for band in bands_list:
            datasets = []
            for cropped, profile in cropped_bands_by_type[band]:
                cropped_epsg, profile_epsg = self.ensure_array_in_epsg(cropped, profile, final_crs)
                profile_epsg.update({
                    'driver': 'GTiff',
                    'dtype': 'uint16',
                    'compress': None,
                    'tiled': False,
                    'nodata': 0
                })
                memfile = rasterio.io.MemoryFile()
                with memfile.open(**profile_epsg) as tmp:
                    tmp.write(cropped_epsg, 1)
                datasets.append(memfile.open())

            merged, out_transform = merge(datasets, method='first')
            merged_bands.append(merged.squeeze())

        # Pega profile de um dos arquivos e atualiza
        if final_profile is None:
            final_profile = datasets[0].profile.copy()
            final_profile.update({
                "height": merged.shape[1],
                "width": merged.shape[2],
                "transform": out_transform,
                "count": 1
            })

            # Fecha arquivos temporários

        return (*merged_bands, final_profile)


    def create_rgb(self, b02_array, b03_array, b04_array, reference_profile=None, output_path=None):
        """
        Cria uma imagem RGB a partir das bandas B02 (Blue), B03 (Green) e B04 (Red) do Sentinel-2.
        
        Parâmetros:
        - b02_array (np.ndarray): Banda 2 (azul)
        - b03_array (np.ndarray): Banda 3 (verde)
        - b04_array (np.ndarray): Banda 4 (vermelho)
        - reference_profile (dict): Perfil de metadados de referência de uma das bandas originais
        - output_path (str, opcional): Caminho para salvar o arquivo. Se None, apenas retorna o array.
        
        Retorna:
        - rgb_array (np.ndarray): Array RGB com shape (3, height, width)
        """

        # Verifica se as dimensões batem
        if not (b02_array.shape == b03_array.shape == b04_array.shape):
            raise ValueError("As bandas devem ter as mesmas dimensões")

        # Cria array RGB no formato (3, altura, largura)
        rgb_array = np.stack([b04_array, b03_array, b02_array], axis=0)  # (R, G, B)

        # Atualiza o perfil para refletir 3 bandas
        profile = reference_profile.copy()
        profile.update({
            'count': 3,
            'dtype': rgb_array.dtype,
            'driver': 'GTiff'
        })

        # Salva ou retorna
        if output_path:
            with rasterio.open(output_path, 'w', **profile) as dst:
                dst.write(rgb_array)
        return rgb_array

    def create_ndvi(self, b04_array, b08_array, reference_profile, output_path=None):
        """
        Calcula o NDVI (Normalized Difference Vegetation Index) usando as bandas B04 (Red) e B08 (NIR) do Sentinel-2.

        Parâmetros:
        - b04_array (np.ndarray): Banda 4 (vermelho)
        - b08_array (np.ndarray): Banda 8 (infravermelho próximo)
        - reference_profile (dict): Perfil de metadados de referência de uma das bandas originais
        - output_path (str, opcional): Caminho para salvar o arquivo. Se None, apenas retorna o array.

        Retorna:
        - ndvi (np.ndarray): Array 2D com valores de NDVI
        """

        # Verifica se as dimensões batem
        if b04_array.shape != b08_array.shape:
            raise ValueError("As bandas devem ter as mesmas dimensões")

        # Converte para float32 para evitar problemas de divisão inteira
        b04 = b04_array.astype('float32')
        b08 = b08_array.astype('float32')

        # NDVI = (NIR - Red) / (NIR + Red)
        denominator = b08 + b04
        with np.errstate(divide='ignore', invalid='ignore'):
            ndvi = (b08 - b04) / denominator
            ndvi[np.isnan(ndvi)] = -9999  # valor nodata

        # Atualiza o perfil
        profile = reference_profile.copy()
        profile.update({
            'count': 1,
            'dtype': 'float32',
            'driver': 'GTiff',
            'nodata': -9999
        })

        # Salva ou retorna
        if output_path:
            with rasterio.open(output_path, 'w', **profile) as dst:
                dst.write(ndvi, 1)
        return ndvi

    def create_evi(self, b02_array, b04_array, b08_array, reference_profile, output_path=None):
        """
        Calcula o EVI (Enhanced Vegetation Index) usando as bandas B02 (Blue), B04 (Red) e B08 (NIR) do Sentinel-2.

        Parâmetros:
        - b02_array (np.ndarray): Banda 2 (azul)
        - b04_array (np.ndarray): Banda 4 (vermelho)
        - b08_array (np.ndarray): Banda 8 (infravermelho próximo)
        - reference_profile (dict): Perfil de metadados de referência de uma das bandas originais
        - output_path (str, opcional): Caminho para salvar o arquivo. Se None, apenas retorna o array.

        Retorna:
        - evi (np.ndarray): Array 2D com valores de EVI
        """

        # Verifica se as dimensões batem
        if not (b02_array.shape == b04_array.shape == b08_array.shape):
            raise ValueError("As bandas devem ter as mesmas dimensões")

        # Converte para float32
        b02 = b02_array.astype('float32') / 10000 # Blue
        b04 = b04_array.astype('float32') / 10000 # Red
        b08 = b08_array.astype('float32') / 10000  # NIR

        # Parâmetros do EVI
        G = 2.5
        C1 = 6.0
        C2 = 7.5
        L = 1.0

        # Fórmula do EVI
        denominator = b08 + C1 * b04 - C2 * b02 + L
        with np.errstate(divide='ignore', invalid='ignore'):
            evi = G * ((b08 - b04) / denominator)
            evi[np.isnan(evi)] = -9999  # valor nodata

        # Atualiza o perfil
        profile = reference_profile.copy()
        profile.update({
            'count': 1,
            'dtype': 'float32',
            'driver': 'GTiff',
            'nodata': -9999
        })

        # Salva ou retorna
        if output_path:
            with rasterio.open(output_path, 'w', **profile) as dst:
                dst.write(evi, 1)
        return evi

if __name__ == '__main__':
    import rasterio
    b02 = '/home/ppz/Documentos/coding/izisat/auxiliary/Sentinel-2/2025/04/06/T22KGD/L2A/10m/T22KGD_20250406T133211_B02_10m.jp2'
    b03 = '/home/ppz/Documentos/coding/izisat/auxiliary/Sentinel-2/2025/04/06/T22KGD/L2A/10m/T22KGD_20250406T133211_B03_10m.jp2'
    b04 = '/home/ppz/Documentos/coding/izisat/auxiliary/Sentinel-2/2025/04/06/T22KGD/L2A/10m/T22KGD_20250406T133211_B04_10m.jp2'
    b08 = '/home/ppz/Documentos/coding/izisat/auxiliary/Sentinel-2/2025/04/06/T22KGD/L2A/10m/T22KGD_20250406T133211_B08_10m.jp2'
    
    # Lê as bandas como arrays e pega o perfil de uma delas
    with rasterio.open(b02) as b2_src:
        b02 = b2_src.read(1)
        profile = b2_src.profile

    with rasterio.open(b03) as b3_src:
        b03 = b3_src.read(1)

    with rasterio.open(b04) as b4_src:
        b04 = b4_src.read(1)
    
    with rasterio.open(b08) as b8_src:
        b08 = b8_src.read(1)

    # Chama a função
    instance = RasterProcessing()
    rgb_array = instance.create_rgb(b02, b03, b04, reference_profile=profile, output_path='rgb_composite.tif')
    ndvi_array = instance.create_ndvi(b04, b08, profile, output_path='ndvi_composite.tif')
